import math
import random
import re
import unidecode

'''
The main function below will train one Markov model for each training dataset
'''
TRAINING_DATASETS = ["Count of Monte Cristo.txt", "Secret Garden.txt", "Scarlet Letter.txt",
                     "Pride and Prejudice.txt", "Green Eggs and Ham.txt"]

'''
This is the number of sequences that will be generated by each trained model
'''
NUM_SEQS_TO_GENERATE = 10

'''
The main function below will compute probabilities from each trained Markov model (see note for TRAINING_DATASETS)
for each test dataset
'''
TEST_DATASETS = ["A Little Princess.txt", "Sense and Sensibility.txt", "The House of Seven Gables.txt",
                 "The Three Musketeers.txt"]

'''
This is where you can adjust the Markov order that is used to train all models in the main function
'''
MARKOV_ORDER = 1

'''
This variable will be set in the main function and is to be used to calculate the probability of out-of-vocabulary (OOV)
tokens when computing the probability of a sequence in which a token is found that was not encountered during training
'''
VOCAB_SIZE = None


class MarkovModel:

    def __init__(self, training_sentences, order=1):
        """
        a dictionary that maps a tuple t (whose length is the Markov order) to the probability that
        t starts a sentence

        It is recommended that you use this datastructure initially to tally counts during training and then
        convert these counts to probabilities when you normalize after training
        """
        self.start_probabilities = {}

        """
        a 2D dictionary that maps a tuple s (prev token) to a tuple t (next token) to the probability that
        t follows s in a sentence (use None as the second key to represent an end of sequence token)

        It is recommended that you use this datastructure initially to tally counts during training and then
        convert these counts to probabilities when you normalize after training
        """
        self.trans_probabilities = {}

        """"
        this is a class variable that represents the Markov order of the model which you can reference in class
        functions
        """
        self.order = order

        '''Here is where you initialize the model (i.e., train parameters) from the input training sentences'''

        # TODO: implement this constructor

    def generate_sequence(self):
        """
        probabilistically generate a sequence of words and return it as a string

        :return: a generated string
        """
        seq = []

        # TODO: implement this function

        return seq

    def compute_log_probability(self, test_sentences):
        """
        use the trained model to compute the probability of
        the input text (a list of lists (sentences) of strings (words)

        :param test_sentences: the text for which to compute the probability according to this model
        :return: the computed probability of the input text
        """

        log_probability = 0.0

        # TODO: implement this function

        return log_probability


def sample(distribution):
    """
        Given a dictionary that maps a tuple t to a probability, this function samples and returns t according to
        its probability

    :param distribution: The dictionary from which to sample
    :return: a tuple
    """
    prob = random.random()
    prob_sum = 0.0

    for key, value in distribution.items():
        prob_sum += value
        if prob_sum >= prob:
            return key

    return key


def parse_file(filename):
    """
    Takes in the name of a file which is parsed and the parsed input cleaned
    Returns a list of lists of strings where each list of strings
    represents a sentence of words (each word is a string)
    """
    filename_w_path = "./data/" + filename
    with open(filename_w_path, 'r', encoding='utf-8-sig') as file:
        sentences = re.split('[;:!?.]', unidecode.unidecode(file.read()).lower())
        sentences = [[word.strip("[]()_-\'\", \n\t")
                      for word in sentence.strip("[]()_-\'\", \n\t").replace('-', ' ').split()]
                     for sentence in sentences]
        sentences = [sentence for sentence in sentences if sentence != []]
    return sentences


def compute_vocab_size():
    vocab = set()
    for TRAINING_DATASET in TRAINING_DATASETS:
        sentences = parse_file(TRAINING_DATASET)
        for sentence in sentences:
            vocab.update(sentence)

    for TEST_DATASET in TEST_DATASETS:
        sentences = parse_file(TEST_DATASET)
        for sentence in sentences:
            vocab.update(sentence)

    return len(vocab)


if __name__ == '__main__':

    VOCAB_SIZE = compute_vocab_size()

    markov_models = {}
    for TRAINING_DATASET in TRAINING_DATASETS:
        # Train a model from scratch
        print("Training new model on", TRAINING_DATASET, "...")
        train_file_sentences = parse_file(TRAINING_DATASET)
        markov_models[TRAINING_DATASET] = MarkovModel(train_file_sentences, MARKOV_ORDER)

    for TRAINING_DATASET in TRAINING_DATASETS:

        if NUM_SEQS_TO_GENERATE > 0:
            print("\nGenerating", NUM_SEQS_TO_GENERATE, "sequences from the model trained on", TRAINING_DATASET)
            for i in range(NUM_SEQS_TO_GENERATE):
                print("\t", markov_models[TRAINING_DATASET].generate_sequence())

    for TEST_DATASET in TEST_DATASETS:
        test_file_sentences = parse_file(TEST_DATASET)

        print("\nThe log probability of test dataset", TEST_DATASET, "is (less negative means more probable)")
        for model_name, markov_model in markov_models.items():
            probability = markov_model.compute_log_probability(test_file_sentences)
            print("\t", probability, "according to the model trained on", model_name)